#include <stdio.h>
#if defined( sgi )
#include <alloca.h>
#else
#include <stdlib.h>
#endif

#include "mpi.h"



/*@

   Init_Task_Grid_order - Initialize the continuous block data array, 
                          Task_Grid[], according to each element's location 
                          in the global grid.  e.g. Task_Grid[3][2][1],
                          in C convention, will be assigned a number
                          iseed*10^(Ndim+1) + 321 formated according to its
                          type.  The routine checks the inter-process
                          communication of the ghost point region of the
                          task grids

  Unmodified Input Variables :
. Ndim - Dimensionality of the Grid (Integer)
. Glb_Grid_sz - Integer Array, of size Ndim, specifies the maximum grid size 
                of the global grid.  Glb_Grid_sz[ ii ] is the maximum grid size
                in the dimension ii, where the minimum is assumed to be 1.
. task_Grid_endpts - Integer Array, of size 2 x Ndim, of the two end-points'
                     coordinates of the LOCAL task grid.  It is implicitly
                     assumed to be declared as task_Grid_endpts[ 2 ][ Ndim ].
. stride_width - Integer Array, of size Ndim, of the stride width of the task
                 grid.  It is implicitly assumed to be declared as
                 stride_width[ Ndim ].
. iseed - Integer specifies the starting digit in each grid element.
. MD_type - the MPI_Datatype of the data stored in the task grids, Task_Grid.

  Modified Output Variables :
. Task_Grid - Void pointer to the Block data array stored in the task grids.
.n

.N RGX_fortran
.N RGX_minimal_set_of_variables
.N RGX_supported_datatype
.N RGX_yes_included

@*/

void Init_Task_Grid_order(
                           int           Ndim,
                           int           Glb_Grid_sz[],
                           int           task_Grid_endpts[],
                           int           stride_width[],
                           int           iseed,
                           MPI_Datatype  MD_type,

                           void         *Task_Grid
                         )
{

#  if defined( DEBUG ) || defined( DEBUG_stdout )
#  include "debug.c.h"
   int       MD_type_sz;
#  endif

   const int BIG_ftr = 10;

   int      *task_grid_sz;
   int      *bndry_lgths;
   int      *offsets;
   int      *Nplet_idx;
   int      *Nplet_world;
   int       hyper_vol;

   int      *Grid_iptr;
   float    *Grid_fptr;
   double   *Grid_dptr;
   int       ftr, itmp;
   int       idx, idir, ii;



/*  Allocate Memory for local variables  */

   task_grid_sz = ( int* ) alloca( Ndim * sizeof( int ) );
   bndry_lgths  = ( int* ) alloca( Ndim * sizeof( int ) );
   offsets      = ( int* ) alloca( Ndim * sizeof( int ) );
   Nplet_idx    = ( int* ) alloca( Ndim * sizeof( int ) );
   Nplet_world  = ( int* ) alloca( Ndim * sizeof( int ) );

/*  Define the boundary lengths of the task grid  */

   hyper_vol = 1;
   for ( idir = 0; idir < Ndim; idir++ )
   {

      task_grid_sz[ idir ] = task_Grid_endpts[ idir + Ndim ]
                           - task_Grid_endpts[ idir ] + 1;
      bndry_lgths[ idir ]  = task_grid_sz[ idir ]
                           + 2 * stride_width[ idir ];
      hyper_vol           *= bndry_lgths[ idir ];

      /*  offsets[] for Fortran style's implicit index [1..N]  */
      /*
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ] - 1; 
      */

      /*  offsets[] for C style's implicit index [0..N-1]  */
      offsets[ idir ]      = task_Grid_endpts[ idir ]
                           - stride_width[ idir ]; 

   }        /*  endif of { for ( idir = 0; idir < Ndim; idir++ ) }  */

#  if defined( DEBUG ) || defined( DEBUG_stdout )
     ierr_debug = MPI_Comm_rank( MPI_COMM_WORLD, &my_tid_debug );
#  endif

#  if defined( DEBUG ) || defined( DEBUG_stdout )
      MPI_Type_size( MD_type, &MD_type_sz );
      fprintf( stdout, "**** MD_type = %d has %d bytes\n",
               MD_type, MD_type_sz );
#  endif

#  ifdef DEBUG_stdout
      fprintf( stdout, "Task %d's array Task_Grid[] at 1 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif

   switch ( MD_type )
   {
      case MPI_INTEGER:
         Grid_iptr = (int*) Task_Grid;
         break;
      case MPI_REAL:
         Grid_fptr = (float*) Task_Grid;
         break;
      case MPI_DOUBLE_PRECISION:
         Grid_dptr = (double*) Task_Grid;
         break;
      case MPI_COMPLEX:
         Grid_fptr = (float*) Task_Grid;
         break;
      case MPI_DOUBLE_COMPLEX:
         Grid_dptr = (double*) Task_Grid;
         break;
      default:
         fprintf( stderr, __FILE__": Wrong type in Init_Task_Grid()\n" );
         break;
   }

   for ( idx = 0; idx < hyper_vol; idx++ )
   {
      adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

      /*  Nplet_world[ ii ] is in [ 1 .. Nmax ] in both Fortran & C  */
      for ( ii = 0; ii < Ndim; ii++ )
         Nplet_world[ ii ] = Nplet_idx[ ii ] + offsets[ ii ];

      if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )
      {
         itmp = 0;
         ftr = 1;
         for ( ii = Ndim-1; ii >= 0; ii-- )
         {
            itmp += ftr * Nplet_world[ ii ];
            ftr *= BIG_ftr;
         }
         itmp += BIG_ftr * ftr * iseed;

         switch ( MD_type )
         {
            case MPI_INTEGER:
               Grid_iptr[ idx ] = itmp;
               break;
            case MPI_REAL:
               Grid_fptr[ idx ] = (float) itmp;
               break;
            case MPI_DOUBLE_PRECISION:
               Grid_dptr[ idx ] = (double) itmp;
               break;
            case MPI_COMPLEX:
               Grid_fptr[ 2*idx   ] =   (float) itmp;
               Grid_fptr[ 2*idx+1 ] = - (float) itmp;
               break;
            case MPI_DOUBLE_COMPLEX:
               Grid_dptr[ 2*idx   ] =   (double) itmp;
               Grid_dptr[ 2*idx+1 ] = - (double) itmp;
               break;
            default:
               fprintf( stderr, __FILE__
                        ": Unable to handle datatype %d in Init_Task_Grid()\n",
                        MD_type );
               return;
         }
      }  /*  if ( Nplet_in_RealWrd( Ndim, task_Grid_endpts, Nplet_world ) )  */
   }     /*  for ( idx = 0; idx < hyper_vol; idx++ )   */

#  if defined( DEBUG )
      sprintf( filename_debug, filename_debug_fmt, my_tid_debug );
      file_debug = fopen( filename_debug, "a+" );
      fprintf( file_debug, "\n" );
      print_Task_Grid( file_debug, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( file_debug, "\n" );
      fclose( file_debug );
#  endif 

#  if defined( DEBUG_stdout )
      fprintf( stdout, "Task %d's array Task_Grid[] at 2 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif 

}



#ifdef FortranNoUnderscore
#   define init_task_grid_order__  init_task_grid_order
#endif

#ifdef FortranSingleUnderscore
#   define init_task_grid_order__  init_task_grid_order_
#endif

/*  Fortran interface for Init_Task_Grid_order()  */
void init_task_grid_order__(
                             int          *Ndim,
                             int          *Glb_Grid_sz,
                             int          *task_Grid_endpts,
                             int          *stride_width,
                             int          *iseed,
                             MPI_Datatype *MD_type,

                             void         *Task_Grid
                           )
{
   Init_Task_Grid_order( *Ndim, Glb_Grid_sz, task_Grid_endpts, stride_width,
                         *iseed, *MD_type, Task_Grid );
}

