#include <stdio.h>
#if defined( sgi )
#include <alloca.h>
#else
#include <stdlib.h>
#endif

#include "mpi.h"

/*
    The 1st routine to be called besides MPI_Init() which is called in Main().
    This routine returns No. of tasks, No. of tasks per dim.
    and the newly formed Communicator in the Cartesian Grids System
*/

/*@

    RGX_Create_Task_GridS - Creates Regular Sub-Grid, one for each process.
                            The first routine to be called after MPI_Init(),

  Unmodified Input Variables :
. World_Comm - Input Communicator containing all processors for the grid 
               (Usually MPI_COMM_WORLD)
. Ndim - Dimensionality of the Grid (Integer)

  Modified Output Variables :
. N_task_ptr - Pointer to Number of Tasks in the Communicator
. N_tasks_1dim - Integer Array, of size Ndim, of Number of Tasks Per Dimension 
                 in the Grid.  It is implicitly assumed to be 
                 N_tasks_1dim[ Ndim ].
. Cart_Grids_Comm_ptr - Pointer to the Cartesian Grid Communicator
. returned value - returns MPI_SUCCESS if completed successfully
.n

.N RGX_fortran
.N RGX_minimal_set_of_variables
.N RGX_supported_datatype
.N RGX_yes_included

@*/

int RGX_Create_Task_GridS(
                           MPI_Comm  World_Comm,
                           int       Ndim,

                           int      *N_task_ptr,
                           int       N_tasks_1dim[],
                           MPI_Comm *Cart_Grids_Comm_ptr
                         )
{
#ifdef DEBUG_stdout
#include "debug.c.h"
#endif

#   define       line_Max  80

    /*  Local variables  */
    const int    master_tid = 0;
    char         local_node_name[ line_Max ];
    int          local_name_lgth;
    int         *periods;
    int          ierr;
    int          idir;
    int          ii;



    /*  Find the No. of Tasks involved in the World_Comm  */

    ierr = MPI_Comm_size( World_Comm, N_task_ptr );
    if ( ierr != MPI_SUCCESS )
    {
       fprintf( stderr, __FILE__": MPI_Comm_size() fails with ierr = %d\n",
                        ierr );
       return ierr;
    }

    /*  Compute the No. of tasks per dim. i.e. N_tasks_1dim[ 0..Ndim-1 ]  */
    /*      N_tasks_1dim[ 0..Ndim-1] HAS TO BE initialized,
            otherwise MPI_Dims_create() will BOMB OUT in MPICH !!!        */

    for ( idir = 0; idir < Ndim; idir++ )   N_tasks_1dim[ idir ] = 0;
    ierr = MPI_Dims_create( *N_task_ptr, Ndim, N_tasks_1dim );


    /*  Assume __PERIODIC__ Boundary Condition on the Global Grid  */

    periods = ( int * ) alloca( Ndim * sizeof( int ) );
    for ( idir = 0; idir < Ndim; idir++ )    periods[ idir ] = 1;


    /*  Create the Cartesian Regular Grids on the given No. of Nodes/Tasks  */

    ierr = MPI_Cart_create( World_Comm, Ndim, N_tasks_1dim, periods,
                            1, Cart_Grids_Comm_ptr );
    if ( ierr != MPI_SUCCESS )
    {
       fprintf( stderr, __FILE__": MPI_Cart_create() fails with ierr = %d\n",
                        ierr );
       return ierr;
    }


#   ifdef DEBUG_stdout
       ierr = MPI_Comm_rank( *Cart_Grids_Comm_ptr, &my_tid_debug );
       ierr = MPI_Get_processor_name( local_node_name, &local_name_lgth );
       fprintf( stdout, "Task %d among { 0 .. %d } is alive on %s\n",
                        my_tid_debug, *N_task_ptr-1, local_node_name );
       if ( my_tid_debug == master_tid )
       {
          fprintf( stdout, "\n Master : No. of Tasks per dim. = {" );
          for ( ii_dbg = 0; ii_dbg < Ndim; ii_dbg++ )
             fprintf( stdout, " %d", N_tasks_1dim[ ii_dbg ] );
          fprintf( stdout, " }\n" );
       }
#   endif

    return MPI_SUCCESS;
}



#ifdef FortranNoUnderscore
#  define rgx_create_task_grids__  rgx_create_task_grids
#endif

#ifdef FortranSingleUnderscore
#  define rgx_create_task_grids__  rgx_create_task_grids_
#endif

/*  Fortran Interface for RGX_Create_Task_GridS()  */
void rgx_create_task_grids__(
                              MPI_Comm *World_Comm,
                              int      *Ndim,

                              int      *N_task,
                              int      *N_tasks_1dim,
                              MPI_Comm *Cart_Grids_Comm,
                              int      *ierr
                            )
{
   *ierr = RGX_Create_Task_GridS( *World_Comm, *Ndim,
                                  N_task, N_tasks_1dim, Cart_Grids_Comm );
}
